const appendOnce = (parent, element, cssClass) => {
  const selector = element + '.' + cssClass
  const child = parent.selectAll(selector).data([0])
  child.enter().append(element).attr('class', cssClass)
  return parent.select(selector)
}

const getOnePixel = (node, width) => {
  const newWidth = node.node().getBoundingClientRect().width
  return width / parseInt(newWidth)
}

const updateSizes = (d3, node, staticNode, defsNode, data, defs, width, height, id) => {
  const onePixel = getOnePixel(node, width)

  const pattern = d3.select('#pattern-stripe')
  if (pattern) {
    pattern
      .attr('height', onePixel * 16)
      .attr('width', onePixel * 16)
    const rects = pattern.selectAll('rect')
    rects
      .attr('height', onePixel * 16)
      .attr('width', onePixel * 8)
      .attr('x', (_, i) => i === 1 ? onePixel * 8 : 0)
  }

  const plus = node.select('.plus')
  if (plus) {
    plus.attr('transform', `translate(${width - 35 * onePixel} ${5 * onePixel}) scale(${onePixel * 1.5})`)
  }
  const minus = node.select('.minus')
  if (minus) {
    minus.attr('transform', `translate(${width - 35 * onePixel} ${37 * onePixel}) scale(${onePixel * 1.5})`)
  }

  if (defs.legend) {
    addLegend(d3, staticNode, defsNode, data, defs, width, height, id, onePixel)
  }
}
const screenSize = (d3, node, staticNode, defsNode, data, defs, width, height, id) => {
  d3.select(window).on(`resize.id${id}`, function() {
    updateSizes(d3, node, staticNode, defsNode, data, defs, width, height, id)
  })
  updateSizes(d3, node, staticNode, defsNode, data, defs, width, height, id)
}

const addGradient = (d3, defs, colors, nr, id) => {
  const colorScale = d3
    .scaleQuantize()
    .range(colors)
  appendOnce(defs, 'linearGradient', `gradient-${nr}-${id}`)
    .attr('id', `gradient-${nr}-${id}`)
    .selectAll('stop')
    .data(colorScale.range())
    .enter()
    .append('stop')
    .attr('offset', (d, i) => { return i / (colorScale.range().length - 1) })
    .attr('stop-color', d => { return d })
}

const darkestHexColor = (colors) => {
  return '#' + colors.map(o => {
    const color = o
      .replace('#', '') // remove hash
      .replace(/^(.)(.)(.)$/, '$1$1$2$2$3$3') // 3 to 6
      .replace(/^(.{6})$/, '$1ff') // 6 to 8
    return parseInt(color, 16)
  }).sort((a, b) => a - b)[0]
    .toString(16) // back to hex again
    .padStart(8, '0') // fill up with zeros
    .replace(/ff$/, '') // remove ff on end
}

const addLegend = (d3, node, defsNode, data, defs, width, height, id, onePixel) => {
  let i = 0
  for (const range of defs.ranges) {
    addGradient(d3, defsNode, range.colors, id, i)
    i++
  }

  const fontSize = onePixel * 16

  const margin = 2
  const padding = 5
  const textHeight = fontSize
  const barHeight = 12 * onePixel

  const outerWidth = width - margin * 2
  const outerHeight = padding + textHeight + padding + barHeight + padding + textHeight + padding
  const innerWidth = outerWidth - padding * 2

  const top = height - outerHeight - margin
  const textBottom = textHeight + padding
  const barTop = textBottom + padding
  const left = margin
  const tickTop = barTop + barHeight + padding

  const legend = appendOnce(node, 'g', 'legend')
    .attr('transform', `translate(${left},${top})`)

  // bg
  appendOnce(legend, 'rect', 'bg')
    .attr('fill', '#ffffffdd')
    .attr('width', outerWidth)
    .attr('height', outerHeight)

  const hasNullValue = data.map(defs.value).find(o => o === null) !== undefined
  const hasStripedValue = defs.striped && data.map(defs.striped).find(o => o) !== undefined

  // null
  let specialBarsWidth = 0
  if (hasNullValue) {
    const nullBar = appendOnce(legend, 'rect', 'null')
      .attr('fill', '#000')
      .attr('y', barTop)
      .attr('height', barHeight).attr('stroke', '#fff')
      .attr('stroke-width', onePixel * 2)
    const nullLabel = appendOnce(legend, 'text', `null-label`)
      .attr('fill', '#000')
      .attr('font-size', fontSize)
      .attr('font-weight', 'normal')
      .attr('dominant-baseline', !hasStripedValue ? 'auto' : 'hanging')
      .text(() => onePixel > 1 ? 'k.A.' : 'keine Angaben')
      .attr('y', !hasStripedValue ? textBottom : tickTop)
    const w = nullLabel.node().getBBox().width + 4
    specialBarsWidth += w
    nullBar
      .attr('width', w)
      .attr('x', innerWidth - specialBarsWidth)
    nullLabel
      .attr('x', innerWidth - specialBarsWidth + 2)
  }

  // stripes
  if (hasStripedValue) {
    addGradient(d3, defsNode, defs.ranges.map(o => o.colors).flat(), id, 'stripe')
    const stripedBar = appendOnce(legend, 'rect', 'stripes')
      .attr('fill', `url(#gradient-${id}-stripe)`)
      .attr('mask', `url(#mask-stripe-${id})`)
      .attr('y', barTop)
      .attr('height', barHeight)
      .attr('stroke', '#fff')
      .attr('stroke-width', onePixel * 2)
    const stripedLabel = appendOnce(legend, 'text', `stripes-label`)
      .attr('fill', '#000')
      .attr('font-size', fontSize)
      .attr('font-weight', 'normal')
      .text(() => onePixel > 1 ? 'n.a.' : 'nicht abgeschlossen')
      .attr('y', textBottom)
    const w = stripedLabel.node().getBBox().width + 4
    specialBarsWidth += w
    stripedBar
      .attr('width', w)
      .attr('x', innerWidth - specialBarsWidth)
    stripedLabel
      .attr('x', innerWidth - specialBarsWidth + 2)
  }

  const rangesWidth = innerWidth - specialBarsWidth
  const barWidth = rangesWidth / defs.ranges.length

  i = 0
  for (const range of defs.ranges) {
    const barLeft = i * barWidth + left
    // bar
    appendOnce(legend, 'rect', `gradient-${id}-${i}`)
      .attr('id', `bar-${id}-${i}`)
      .attr('fill', `url(#gradient-${id}-${i})`)
      .attr('x', barLeft)
      .attr('y', barTop)
      .attr('width', barWidth)
      .attr('height', barHeight)
      .attr('stroke', '#fff')
      .attr('stroke-width', onePixel * 2)

    if (range.label) {
      // label
      appendOnce(legend, 'text', `label-${i}`)
        .attr('text-anchor', i === 0 ? 'start' : (i === defs.ranges.length - 1 ? 'end' : 'middle'))
        .attr('fill', range.labelColor || darkestHexColor(range.colors))
        .attr('font-size', fontSize)
        .text(range.label)
        .attr('y', textBottom)
        .attr('x', i === 0 ? barLeft : (i === defs.ranges.length - 1 ? barLeft + barWidth - 2 : barLeft + barWidth / 2))
    }

    function values () {
      return range.ticks || range.values
    }

    // ticks
    function shiftLabelX (d, index) {
      const tickValues = values()
      // prev ticks
      if (i > 0) {
        const prevRange = defs.ranges[i - 1]
        const prevTickValues = (prevRange.ticks || prevRange.values)
        // if last tick from prev range the same value as first tick of actual range => center it
        if (index === 0 && prevTickValues && prevTickValues[prevTickValues.length - 1] === tickValues[0]) {
          return barLeft - this.getBBox().width * 0.5
        }
      }
      const shift = 1 / (tickValues.length - 1) * index
      return barLeft + margin + (barWidth - 2 * margin) * shift - this.getBBox().width * shift
    }
    function labelOpacity (d, index) {
      const tickValues = values()
      // next ticks
      if (i < defs.ranges.length - 1) {
        const nextRange = defs.ranges[i + 1]
        const nextTickValues = (nextRange.ticks || nextRange.values)
        // if first tick from next range the same value as last tick of actual range => hide it
        if (index === tickValues.length - 1 && nextTickValues[0] === tickValues[tickValues.length - 1]) {
          return 0
        }
      }
      return 1
    }

    const ticks = legend.selectAll(`.tick-${i}`).data(values)
    ticks.join(
      enter => enter
        .append('text')
        .attr('class', `tick-${i}`)
        .attr('font-weight', 'normal')
        .attr('dominant-baseline', 'hanging')
        .attr('fill', '#000')
        .attr('opacity', labelOpacity)
        .attr('font-size', fontSize)
        .text(d => d.toFixed(1).replace(/.0$/, ''))
        .attr('x', shiftLabelX)
        .attr('y', tickTop),
      update => update
        .attr('font-size', fontSize)
        .attr('opacity', labelOpacity)
        .text(d => d.toFixed(1).replace(/.0$/, ''))
        .attr('x', shiftLabelX)
        .attr('y', tickTop),
    )

    i++
  }

  // selector
  appendOnce(legend, 'rect', `selector`)
    .attr('id', `selector-${id}`)
    .attr('data', rangesWidth)
    .attr('fill', '#000')
    .attr('x', rangesWidth / 2)
    .attr('y', barTop)
    .attr('width', onePixel * 5)
    .attr('height', barHeight)
    .attr('stroke', '#fff')
    .attr('stroke-width', onePixel)
    .attr('opacity', 0)

  return top
}

const circle = (radius, center) => {
  return `M ${center.x - radius}, ${center.y} a ${radius},${radius} 0 1,0 ${radius * 2},0 a ${radius},${radius} 0 1,0 -${radius * 2},0`
}

const addTooltip = (root) => {
  const tooltip = root.append('div')
  root.attr('style', 'position:relative;')
  tooltip.attr('style', 'opacity:0; z-index:9999; background-color:white; padding:16px; pointer-events:none; position:absolute; box-shadow:0 0 16px #00000066;')
  return tooltip
}

const svgDataURL = (svg) => {
  const svgAsXML = (new XMLSerializer()).serializeToString(svg)
  return 'data:image/svg+xml,' + encodeURIComponent(svgAsXML)
}

const addStripePattern = (defs, id) => {
  const size = 16
  const pattern = appendOnce(defs, 'pattern', 'pattern-stripe')
    .attr('id', `pattern-stripe-${id}`)
    .attr('width', size)
    .attr('height', size)
    .attr('patternUnits', 'userSpaceOnUse')
    .attr('patternTransform', 'rotate(-30)')
  appendOnce(pattern, 'rect', 'one')
    .attr('width', size / 2)
    .attr('height', size)
    .attr('fill', '#bbb')
  appendOnce(pattern, 'rect', 'two')
    .attr('x', size / 2)
    .attr('width', size / 2)
    .attr('height', size)
    .attr('fill', '#fff')
  const mask = appendOnce(defs, 'mask', 'mask')
    .attr('id', `mask-stripe-${id}`)
  appendOnce(mask, 'rect', 'bounds')
    .attr('x', '0')
    .attr('y', '0')
    .attr('width', '100%')
    .attr('height', '100%')
    .attr('fill', `url(#pattern-stripe-${id})`)
}

const downloadDataURL = (svg, name) => {
  const dl = document.createElement('a')
  document.body.appendChild(dl) // This line makes it work in Firefox.
  dl.setAttribute('href', svgDataURL(svg))
  dl.setAttribute('download', `${name}.svg`)
  dl.click()
}

const addZoomIn = (d3, node, callback, x = 0, y = 0, size = 24) => {
  const padding = 6
  // plus
  const plus = appendOnce(node, 'g', 'plus')
  plus.attr('transform', 'translate(-100, -100)')
  plus.append('rect')
    .attr('width', size)
    .attr('height', size)
    .attr('fill', '#ccc')
    .on('mouseover', function (event, d) {
      const item = d3.select(this)
      item.attr('style', 'filter:contrast(50%) brightness(140%); cursor:pointer;')
    })
    .on('mouseout', function (event, d) {
      const item = d3.select(this)
      item.attr('style', '')
    })
    .on('click', function (event) {
      event.stopPropagation()
      callback()
    })

  plus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', padding)
    .attr('y', (size - 1) / 2)
    .attr('width', size - 2 * padding)
    .attr('height', 1)
    .attr('fill', '#000')

  plus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', (size - 1) / 2)
    .attr('y', padding)
    .attr('width', 1)
    .attr('height', size - 2 * padding)
    .attr('fill', '#000')
}

const addZoomOut = (d3, node, callback, x = 0, y = 0, size = 24) => {
  const padding = 6
  // minus
  const minus = appendOnce(node, 'g', 'minus')
  minus.attr('transform', 'translate(-100, -100)')
  minus.append('rect')
    .attr('width', size)
    .attr('height', size)
    .attr('fill', '#ccc')
    .on('mouseover', function (event, d) {
      const item = d3.select(this)
      item.attr('style', 'filter:contrast(50%) brightness(140%); cursor:pointer;')
    })
    .on('mouseout', function (event, d) {
      const item = d3.select(this)
      item.attr('style', '')
    })
    .on('click', function (event) {
      event.stopPropagation()
      callback()
    })

  minus.append('rect')
    .attr('pointer-events', 'none')
    .attr('x', padding)
    .attr('y', (size - 1) / 2)
    .attr('width', size - 2 * padding)
    .attr('height', 1)
    .attr('fill', '#000')
}

const map = async (d3, target, data = [], defs = {}, selectedId = null, width = 600, height = 600) => {
  const id = d3.selectAll('.svg').nodes().length
  let selectedCallback = null
  const root = d3.select(target)

  const svg = appendOnce(root, 'svg', 'svg')
    .attr('viewBox', [0, 0, width, height])
    .attr('font-family', 'Arial, Helvetica, sans-serif')
    .attr('font-weight', 'bold')

  let zoom = null
  const zoomNode = appendOnce(svg, 'g', 'zoom-area')
  const staticNode = appendOnce(svg, 'g', 'static-area')

  if (defs.zoomable) {
    addZoomIn(d3, staticNode, () => {
      svg.transition().call(zoom.scaleBy, 2)
    }, width - 25, 5, 20)
    addZoomOut(d3, staticNode, () => {
      svg.transition().call(zoom.scaleBy, 0.5)
    }, width - 25, 27, 20)
  }

  const bounds = {
    type: 'FeatureCollection',
    features: []
  }

  if (!defs.url) {
    console.error('url prop is requied data item!')
  }
  const geojson = await d3.json(defs.url)
  const itemIds = data.map(o => defs.itemId(o))
  bounds.features.push(...geojson.features.filter(o => defs.all || itemIds.includes(defs.mapId(o.properties))))

  let bg = null
  if (defs.bg) {
    bg = await d3.json(defs.bg)
    bounds.features.push(...bg.features)
  }

  const defsNode = appendOnce(svg, 'defs', 'defs')
  if (defs.striped) {
    addStripePattern(defsNode, id)
  }

  let top = height
  if (defs.legend) {
    top = height - 70 * getOnePixel(svg, width)
  }

  const projection = d3.geoIdentity().reflectY(true).fitSize([width, top], bounds)
  const path = d3.geoPath().projection(projection)
  const geometry = d => {
    const found = bounds.features.find(o => defs.mapId(o.properties) === defs.itemId(d))
    if (!found) {
      return null
    }
    return path(found.geometry)
  }
  const center = d => {
    const found = bounds.features.find(o => defs.mapId(o.properties) === defs.itemId(d))
    if (!found) {
      return { x: -Infinity, y: -Infinity }
    }
    const center = path.centroid(found.geometry)
    if (!center) {
      return { x: -Infinity, y: -Infinity }
    }
    return { x: center[0], y: center[1] }
  }
  const bubbleGeometry = (d) => {
    return defs.bubbleGeometry ? defs.bubbleGeometry(defs.bubble(d), center(d)) : circle(defs.bubble(d), center(d))
  }
  const color = d3
    .scaleLinear()
    .domain(defs.ranges.map(o => o.values).flat())
    .range(defs.ranges.map(o => o.colors).flat())
    .unknown('#000')

  const bgNode = appendOnce(zoomNode, 'g', 'bg')
  if (defs.bg) {
    bgNode.selectAll('path.bg')
      .data(bg.features)
      .enter()
      .append('path')
      .attr('class', 'bg')
      .attr('d', path)
      .attr('fill', '#ddd')
      .attr('stroke', '#fff')
      .attr('stroke-width', 2)
  }

  if (defs.bubble) {
    bgNode.selectAll('path.forms')
      .data(geojson.features.filter(o => defs.all || itemIds.includes(defs.mapId(o.properties))))
      .enter()
      .append('path')
      .attr('class', 'forms')
      .attr('d', path)
      .attr('fill', '#eee')
      .attr('stroke', '#fff')
      .attr('stroke-width', 2)
  }

  let tooltip = null
  if (defs.tooltip) {
    tooltip = addTooltip(root)
  }

  const mouseOver = function (event, d) {
    const item = d3.select(this)
    item.attr('stroke', '#000')
    item.attr('fill', 'rgba(255,255,255,0.3)')

    const value = defs.value(d)
    if (value !== null) {
      const selector = svg.select(`#selector-${id}`)
      if (!selector.empty()) {
        selector.attr('opacity', 1)

        const range = defs.ranges.find(o => o.values[0] <= value && value <= o.values[o.values.length - 1])
        const index = defs.ranges.indexOf(range)

        const bar = svg.select(`#bar-${id}-${index}`)
        if (!bar.empty()) {
          const widths = [...Array(range.values.length)]
          .map((_, i) => 1 / (range.values.length - 1) * i) // interpolate 0 - 1
          .map(o => o * parseFloat(bar.attr('width')))
          const x = d3
            .scaleLinear()
            .domain(range.values)
            .range(widths)
          const left = parseFloat(bar.attr('x')) + x(value) - parseFloat(selector.attr('width')) / 2
          selector.transition(100).attr('x', left)
        }
      }
    }

    // item.raise()
    if (defs.tooltip) {
      tooltip
        .html(defs.tooltip(d))
        .style('opacity', 1)
      mouseMove(event, d)
    }
  }

  const mouseMove = function (event, d) {
    if (defs.tooltip) {
      const pos = root.node().getBoundingClientRect()
      const center = pos.x + pos.width / 2
      let x = event.clientX - pos.x + 20
      if (event.clientX > center) {
        x = event.clientX - pos.x - 20 - tooltip.node().getBoundingClientRect().width
      }
      const y = event.clientY - pos.y
      tooltip
        .style('left', x + 'px')
        .style('top', y + 'px')
    }
  }

  const mouseOut = function (event, d) {
    const item = d3.select(this)
    item.attr('stroke', 'transparent')
    item.attr('fill', null)

    const selector = svg.select(`#selector-${id}`)
    selector.attr('opacity', 0)

    if (defs.tooltip) {
      tooltip
        .style('opacity', 0)
    }
  }

  const zoomToItem = nr => {
    const item = d3.select(`#item-${id}-${nr}`)
    const bBox = item.node().getBBox()
    svg.transition().duration(750).call(
      zoom.transform,
      d3.zoomIdentity
        .translate(width / 2, height / 2)
        .scale(0.6 * Math.min(width / bBox.width, height / bBox.height))
        .translate(-bBox.x - bBox.width / 2, -bBox.y - bBox.height / 2),
    )
  }

  const click = function (event, d) {
    event.stopPropagation()
    const newId = defs.itemId(d)
    selectedId = selectedId === newId ? null : newId

    if (defs.selectable) {
      if (selectedId) {
        if (defs.autofocus) {
          zoomToItem(selectedId)
        }
        if (selectedCallback) {
          selectedCallback(selectedId)
        }
      } else {
        if (defs.unselectable) {
          reset()
          if (selectedCallback) {
            selectedCallback(null)
          }
        }
      }
    }
  }

  const colorNode = appendOnce(zoomNode, 'g', 'color')
    .attr('stroke', '#fff')
    .attr('stroke-width', 1)
    .attr('pointer-events', 'none')
  let hoverLineNode = null
  let selectionNode = null
  let selectionShadow = null
  if (defs.hoverable) {
    if (defs.selectable) {
      selectionShadow = appendOnce(zoomNode, 'g', 'selection-shadow')
        .attr('stroke-linejoin', 'round')
        .attr('pointer-events', 'none')
      selectionNode = appendOnce(zoomNode, 'g', 'selection')
        .attr('pointer-events', 'none')
    }
    hoverLineNode = appendOnce(zoomNode, 'g', 'hover-line')
      .attr('stroke', 'transparent')
      .attr('fill', 'transparent')
      .attr('stroke-width', 1)
  }

  const update = () => {
    if (defs.legend || defs.zoomable) {
      screenSize(d3, svg, staticNode, defsNode, data, defs, width, height, id)
    }

    colorNode.selectAll('path')
      .data([...data].sort((a, b) => defs.bubble ? d3.ascending(defs.bubble(b) || Infinity, defs.bubble(a) || Infinity) : 0), defs.itemId)
      .join(
        enter => enter
          .append('path')
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', i => defs.bubble && defs.value(i) !== null ? bubbleGeometry(i) : geometry(i))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        update => update
          .attr('id', d => `item-${id}-${defs.itemId(d)}`)
          .attr('d', i => defs.bubble && defs.value(i) !== null ? bubbleGeometry(i) : geometry(i))
          .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
          .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
        exit => exit
          .remove()
      )

    // hover
    if (defs.hoverable) {
      const hoverNodes = hoverLineNode.selectAll(`path`)
      .data(data, defs.mapId)
      .join(
        enter => enter
          .append('path')
          .attr('d', geometry),
        update => update
          .attr('d', geometry),
        exit => exit
          .remove()
      )

      hoverNodes
        .on('mouseover', mouseOver)
        .on('mousemove', mouseMove)
        .on('mouseout', mouseOut)

      // select
      if (defs.selectable) {
        hoverNodes
          .on('click', click)
          .style('cursor', 'pointer')
      }
    }

    if (defs.selectable && selectedId) {
      // wait for item is rendered
      setTimeout(() => {
        const item = colorNode.select(`#item-${id}-${selectedId}`)
        const steps = 6
        const size = 2
        const shadows = [...Array(steps)].map((_, i) => { return { id: i } })
        selectionShadow.selectAll(`path`)
          .data(shadows, d => d.id)
          .join(
            enter => enter
              .append('path')
              .attr('d', item.attr('d'))
              .attr('stroke-width', (_, i) => Math.max(0, steps * size - i * size))
              .attr('stroke', (_, i) => i + 1 < steps ? '#00000022' : '#fff'),
            update => update
              .attr('d', item.attr('d'))
              .attr('stroke-width', (_, i) => Math.max(0, steps * size - i * size))
              .attr('stroke', (_, i) => i + 1 < steps ? '#00000022' : '#fff'),
            exit => exit
              .remove()
          )

        selectionNode.selectAll('path')
          .data(data.filter(o => defs.itemId(o) === selectedId), defs.itemId)
          .join(
            enter => enter
              .append('path')
              .attr('d', i => defs.bubble ? bubbleGeometry(i) : geometry(i))
              .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
              .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
            update => update
              .attr('d', i => defs.bubble ? bubbleGeometry(i) : geometry(i))
              .attr('fill', d => defs.color ? defs.color(d, color(defs.value(d))) : color(defs.value(d)))
              .attr('mask', d => defs.striped && defs.striped(d) ? `url(#mask-stripe-${id})` : ''),
            exit => exit
              .remove()
          )
      }, 0)
    }
  }
  update()

  function zoomed ({ transform }) {
    zoomNode.attr('transform', transform)
  }

  zoom = d3.zoom()
    .scaleExtent([1, 4])
    .translateExtent([[0, 0], [width, height]]) // limit drag
    .on('zoom', zoomed)

  if (defs.zoomable) {
    svg.call(zoom) // use to enable drag
  }

  function reset (point) {
    svg.transition().duration(750).call(
      zoom.transform,
      d3.zoomIdentity,
      d3.zoomTransform(svg.node()).invert([width / 2, height / 2])
    )
  }

  return {
    selectedId: (id) => {
      if (id !== null && !data.find(o => defs.itemId(o) === id)) {
        console.error(`Id: ${id} not found in data !`)
        return
      }

      if (selectedId) {
        selectedId = id
        if (defs.autofocus) {
          zoomToItem(selectedId)
        }
      } else {
        selectedId = null
        reset()
      }
      update()
    },
    selectedIdCallback: (callbackFunction) => {
      selectedCallback = callbackFunction
    },
    updateData: (_data) => {
      data = _data
      update()
    },
    updateDefs: (_defs) => {
      defs = _defs
      update()
    },
    update,
    reset,
    zoomIn: () => {
      if (defs.zoomable) {
        svg.transition().call(zoom.scaleBy, 2)
      }
    },
    zoomOut: () => {
      if (defs.zoomable) {
        svg.transition().call(zoom.scaleBy, 0.5)
      }
    },
    download: (name, title) => {
      // title
      appendOnce(svg, 'text', 'title')
        .attr('fill', '#000')
        .attr('alignment-baseline', 'baseline')
        .text(title)
        .attr('font-size', '20px')
        .attr('x', 10)
        .attr('y', 0)

      svg.attr('viewBox', [0, -30, width, height + 30])
      downloadDataURL(svg.node(), name)
      svg.attr('viewBox', [0, 0, width, height])
    }
  }
}

export {
  map
}
